// Converted from http://www.unicode.org/Public/UCD/latest/ucd/Blocks.txt by block_scripts.l
#include <reflex/unicode.h>
void reflex::Unicode::Tables::block_scripts(void)
{
  static const int IsAdlam[] = { 125184, 125279, 0, 0 };
  range["IsAdlam"] = IsAdlam;
  static const int IsAegeanNumbers[] = { 65792, 65855, 0, 0 };
  range["IsAegeanNumbers"] = IsAegeanNumbers;
  static const int IsAhom[] = { 71424, 71503, 0, 0 };
  range["IsAhom"] = IsAhom;
  static const int IsAlchemicalSymbols[] = { 128768, 128895, 0, 0 };
  range["IsAlchemicalSymbols"] = IsAlchemicalSymbols;
  static const int IsAlphabeticPresentationForms[] = { 64256, 64335, 0, 0 };
  range["IsAlphabeticPresentationForms"] = IsAlphabeticPresentationForms;
  static const int IsAnatolianHieroglyphs[] = { 82944, 83583, 0, 0 };
  range["IsAnatolianHieroglyphs"] = IsAnatolianHieroglyphs;
  static const int IsAncientGreekMusicalNotation[] = { 119296, 119375, 0, 0 };
  range["IsAncientGreekMusicalNotation"] = IsAncientGreekMusicalNotation;
  static const int IsAncientGreekNumbers[] = { 65856, 65935, 0, 0 };
  range["IsAncientGreekNumbers"] = IsAncientGreekNumbers;
  static const int IsAncientSymbols[] = { 65936, 65999, 0, 0 };
  range["IsAncientSymbols"] = IsAncientSymbols;
  static const int IsArabic[] = { 1536, 1791, 0, 0 };
  range["IsArabic"] = IsArabic;
  static const int IsArabicExtended_A[] = { 2208, 2303, 0, 0 };
  range["IsArabicExtended-A"] = IsArabicExtended_A;
  static const int IsArabicExtended_B[] = { 2160, 2207, 0, 0 };
  range["IsArabicExtended-B"] = IsArabicExtended_B;
  static const int IsArabicExtended_C[] = { 69312, 69375, 0, 0 };
  range["IsArabicExtended-C"] = IsArabicExtended_C;
  static const int IsArabicMathematicalAlphabeticSymbols[] = { 126464, 126719, 0, 0 };
  range["IsArabicMathematicalAlphabeticSymbols"] = IsArabicMathematicalAlphabeticSymbols;
  static const int IsArabicPresentationForms_A[] = { 64336, 65023, 0, 0 };
  range["IsArabicPresentationForms-A"] = IsArabicPresentationForms_A;
  static const int IsArabicPresentationForms_B[] = { 65136, 65279, 0, 0 };
  range["IsArabicPresentationForms-B"] = IsArabicPresentationForms_B;
  static const int IsArabicSupplement[] = { 1872, 1919, 0, 0 };
  range["IsArabicSupplement"] = IsArabicSupplement;
  static const int IsArmenian[] = { 1328, 1423, 0, 0 };
  range["IsArmenian"] = IsArmenian;
  static const int IsArrows[] = { 8592, 8703, 0, 0 };
  range["IsArrows"] = IsArrows;
  static const int IsAvestan[] = { 68352, 68415, 0, 0 };
  range["IsAvestan"] = IsAvestan;
  static const int IsBalinese[] = { 6912, 7039, 0, 0 };
  range["IsBalinese"] = IsBalinese;
  static const int IsBamum[] = { 42656, 42751, 0, 0 };
  range["IsBamum"] = IsBamum;
  static const int IsBamumSupplement[] = { 92160, 92735, 0, 0 };
  range["IsBamumSupplement"] = IsBamumSupplement;
  static const int IsBasicLatin[] = { 0, 127, 0, 0 };
  range["IsBasicLatin"] = IsBasicLatin;
  static const int IsBassaVah[] = { 92880, 92927, 0, 0 };
  range["IsBassaVah"] = IsBassaVah;
  static const int IsBatak[] = { 7104, 7167, 0, 0 };
  range["IsBatak"] = IsBatak;
  static const int IsBengali[] = { 2432, 2559, 0, 0 };
  range["IsBengali"] = IsBengali;
  static const int IsBhaiksuki[] = { 72704, 72815, 0, 0 };
  range["IsBhaiksuki"] = IsBhaiksuki;
  static const int IsBlockElements[] = { 9600, 9631, 0, 0 };
  range["IsBlockElements"] = IsBlockElements;
  static const int IsBopomofo[] = { 12544, 12591, 0, 0 };
  range["IsBopomofo"] = IsBopomofo;
  static const int IsBopomofoExtended[] = { 12704, 12735, 0, 0 };
  range["IsBopomofoExtended"] = IsBopomofoExtended;
  static const int IsBoxDrawing[] = { 9472, 9599, 0, 0 };
  range["IsBoxDrawing"] = IsBoxDrawing;
  static const int IsBrahmi[] = { 69632, 69759, 0, 0 };
  range["IsBrahmi"] = IsBrahmi;
  static const int IsBraillePatterns[] = { 10240, 10495, 0, 0 };
  range["IsBraillePatterns"] = IsBraillePatterns;
  static const int IsBuginese[] = { 6656, 6687, 0, 0 };
  range["IsBuginese"] = IsBuginese;
  static const int IsBuhid[] = { 5952, 5983, 0, 0 };
  range["IsBuhid"] = IsBuhid;
  static const int IsByzantineMusicalSymbols[] = { 118784, 119039, 0, 0 };
  range["IsByzantineMusicalSymbols"] = IsByzantineMusicalSymbols;
  static const int IsCJKCompatibility[] = { 13056, 13311, 0, 0 };
  range["IsCJKCompatibility"] = IsCJKCompatibility;
  static const int IsCJKCompatibilityForms[] = { 65072, 65103, 0, 0 };
  range["IsCJKCompatibilityForms"] = IsCJKCompatibilityForms;
  static const int IsCJKCompatibilityIdeographs[] = { 63744, 64255, 0, 0 };
  range["IsCJKCompatibilityIdeographs"] = IsCJKCompatibilityIdeographs;
  static const int IsCJKCompatibilityIdeographsSupplement[] = { 194560, 195103, 0, 0 };
  range["IsCJKCompatibilityIdeographsSupplement"] = IsCJKCompatibilityIdeographsSupplement;
  static const int IsCJKRadicalsSupplement[] = { 11904, 12031, 0, 0 };
  range["IsCJKRadicalsSupplement"] = IsCJKRadicalsSupplement;
  static const int IsCJKStrokes[] = { 12736, 12783, 0, 0 };
  range["IsCJKStrokes"] = IsCJKStrokes;
  static const int IsCJKSymbolsandPunctuation[] = { 12288, 12351, 0, 0 };
  range["IsCJKSymbolsandPunctuation"] = IsCJKSymbolsandPunctuation;
  static const int IsCJKUnifiedIdeographs[] = { 19968, 40959, 0, 0 };
  range["IsCJKUnifiedIdeographs"] = IsCJKUnifiedIdeographs;
  static const int IsCJKUnifiedIdeographsExtensionA[] = { 13312, 19903, 0, 0 };
  range["IsCJKUnifiedIdeographsExtensionA"] = IsCJKUnifiedIdeographsExtensionA;
  static const int IsCJKUnifiedIdeographsExtensionB[] = { 131072, 173791, 0, 0 };
  range["IsCJKUnifiedIdeographsExtensionB"] = IsCJKUnifiedIdeographsExtensionB;
  static const int IsCJKUnifiedIdeographsExtensionC[] = { 173824, 177983, 0, 0 };
  range["IsCJKUnifiedIdeographsExtensionC"] = IsCJKUnifiedIdeographsExtensionC;
  static const int IsCJKUnifiedIdeographsExtensionD[] = { 177984, 178207, 0, 0 };
  range["IsCJKUnifiedIdeographsExtensionD"] = IsCJKUnifiedIdeographsExtensionD;
  static const int IsCJKUnifiedIdeographsExtensionE[] = { 178208, 183983, 0, 0 };
  range["IsCJKUnifiedIdeographsExtensionE"] = IsCJKUnifiedIdeographsExtensionE;
  static const int IsCJKUnifiedIdeographsExtensionF[] = { 183984, 191471, 0, 0 };
  range["IsCJKUnifiedIdeographsExtensionF"] = IsCJKUnifiedIdeographsExtensionF;
  static const int IsCJKUnifiedIdeographsExtensionG[] = { 196608, 201551, 0, 0 };
  range["IsCJKUnifiedIdeographsExtensionG"] = IsCJKUnifiedIdeographsExtensionG;
  static const int IsCJKUnifiedIdeographsExtensionH[] = { 201552, 205743, 0, 0 };
  range["IsCJKUnifiedIdeographsExtensionH"] = IsCJKUnifiedIdeographsExtensionH;
  static const int IsCJKUnifiedIdeographsExtensionI[] = { 191472, 192095, 0, 0 };
  range["IsCJKUnifiedIdeographsExtensionI"] = IsCJKUnifiedIdeographsExtensionI;
  static const int IsCarian[] = { 66208, 66271, 0, 0 };
  range["IsCarian"] = IsCarian;
  static const int IsCaucasianAlbanian[] = { 66864, 66927, 0, 0 };
  range["IsCaucasianAlbanian"] = IsCaucasianAlbanian;
  static const int IsChakma[] = { 69888, 69967, 0, 0 };
  range["IsChakma"] = IsChakma;
  static const int IsCham[] = { 43520, 43615, 0, 0 };
  range["IsCham"] = IsCham;
  static const int IsCherokee[] = { 5024, 5119, 0, 0 };
  range["IsCherokee"] = IsCherokee;
  static const int IsCherokeeSupplement[] = { 43888, 43967, 0, 0 };
  range["IsCherokeeSupplement"] = IsCherokeeSupplement;
  static const int IsChessSymbols[] = { 129536, 129647, 0, 0 };
  range["IsChessSymbols"] = IsChessSymbols;
  static const int IsChorasmian[] = { 69552, 69599, 0, 0 };
  range["IsChorasmian"] = IsChorasmian;
  static const int IsCombiningDiacriticalMarks[] = { 768, 879, 0, 0 };
  range["IsCombiningDiacriticalMarks"] = IsCombiningDiacriticalMarks;
  static const int IsCombiningDiacriticalMarksExtended[] = { 6832, 6911, 0, 0 };
  range["IsCombiningDiacriticalMarksExtended"] = IsCombiningDiacriticalMarksExtended;
  static const int IsCombiningDiacriticalMarksSupplement[] = { 7616, 7679, 0, 0 };
  range["IsCombiningDiacriticalMarksSupplement"] = IsCombiningDiacriticalMarksSupplement;
  static const int IsCombiningDiacriticalMarksforSymbols[] = { 8400, 8447, 0, 0 };
  range["IsCombiningDiacriticalMarksforSymbols"] = IsCombiningDiacriticalMarksforSymbols;
  static const int IsCombiningHalfMarks[] = { 65056, 65071, 0, 0 };
  range["IsCombiningHalfMarks"] = IsCombiningHalfMarks;
  static const int IsCommonIndicNumberForms[] = { 43056, 43071, 0, 0 };
  range["IsCommonIndicNumberForms"] = IsCommonIndicNumberForms;
  static const int IsControlPictures[] = { 9216, 9279, 0, 0 };
  range["IsControlPictures"] = IsControlPictures;
  static const int IsCoptic[] = { 11392, 11519, 0, 0 };
  range["IsCoptic"] = IsCoptic;
  static const int IsCopticEpactNumbers[] = { 66272, 66303, 0, 0 };
  range["IsCopticEpactNumbers"] = IsCopticEpactNumbers;
  static const int IsCountingRodNumerals[] = { 119648, 119679, 0, 0 };
  range["IsCountingRodNumerals"] = IsCountingRodNumerals;
  static const int IsCuneiform[] = { 73728, 74751, 0, 0 };
  range["IsCuneiform"] = IsCuneiform;
  static const int IsCuneiformNumbersandPunctuation[] = { 74752, 74879, 0, 0 };
  range["IsCuneiformNumbersandPunctuation"] = IsCuneiformNumbersandPunctuation;
  static const int IsCurrencySymbols[] = { 8352, 8399, 0, 0 };
  range["IsCurrencySymbols"] = IsCurrencySymbols;
  static const int IsCypriotSyllabary[] = { 67584, 67647, 0, 0 };
  range["IsCypriotSyllabary"] = IsCypriotSyllabary;
  static const int IsCypro_Minoan[] = { 77712, 77823, 0, 0 };
  range["IsCypro-Minoan"] = IsCypro_Minoan;
  static const int IsCyrillic[] = { 1024, 1279, 0, 0 };
  range["IsCyrillic"] = IsCyrillic;
  static const int IsCyrillicExtended_A[] = { 11744, 11775, 0, 0 };
  range["IsCyrillicExtended-A"] = IsCyrillicExtended_A;
  static const int IsCyrillicExtended_B[] = { 42560, 42655, 0, 0 };
  range["IsCyrillicExtended-B"] = IsCyrillicExtended_B;
  static const int IsCyrillicExtended_C[] = { 7296, 7311, 0, 0 };
  range["IsCyrillicExtended-C"] = IsCyrillicExtended_C;
  static const int IsCyrillicExtended_D[] = { 122928, 123023, 0, 0 };
  range["IsCyrillicExtended-D"] = IsCyrillicExtended_D;
  static const int IsCyrillicSupplement[] = { 1280, 1327, 0, 0 };
  range["IsCyrillicSupplement"] = IsCyrillicSupplement;
  static const int IsDeseret[] = { 66560, 66639, 0, 0 };
  range["IsDeseret"] = IsDeseret;
  static const int IsDevanagari[] = { 2304, 2431, 0, 0 };
  range["IsDevanagari"] = IsDevanagari;
  static const int IsDevanagariExtended[] = { 43232, 43263, 0, 0 };
  range["IsDevanagariExtended"] = IsDevanagariExtended;
  static const int IsDevanagariExtended_A[] = { 72448, 72543, 0, 0 };
  range["IsDevanagariExtended-A"] = IsDevanagariExtended_A;
  static const int IsDingbats[] = { 9984, 10175, 0, 0 };
  range["IsDingbats"] = IsDingbats;
  static const int IsDivesAkuru[] = { 71936, 72031, 0, 0 };
  range["IsDivesAkuru"] = IsDivesAkuru;
  static const int IsDogra[] = { 71680, 71759, 0, 0 };
  range["IsDogra"] = IsDogra;
  static const int IsDominoTiles[] = { 127024, 127135, 0, 0 };
  range["IsDominoTiles"] = IsDominoTiles;
  static const int IsDuployan[] = { 113664, 113823, 0, 0 };
  range["IsDuployan"] = IsDuployan;
  static const int IsEarlyDynasticCuneiform[] = { 74880, 75087, 0, 0 };
  range["IsEarlyDynasticCuneiform"] = IsEarlyDynasticCuneiform;
  static const int IsEgyptianHieroglyphFormatControls[] = { 78896, 78943, 0, 0 };
  range["IsEgyptianHieroglyphFormatControls"] = IsEgyptianHieroglyphFormatControls;
  static const int IsEgyptianHieroglyphs[] = { 77824, 78895, 0, 0 };
  range["IsEgyptianHieroglyphs"] = IsEgyptianHieroglyphs;
  static const int IsElbasan[] = { 66816, 66863, 0, 0 };
  range["IsElbasan"] = IsElbasan;
  static const int IsElymaic[] = { 69600, 69631, 0, 0 };
  range["IsElymaic"] = IsElymaic;
  static const int IsEmoticons[] = { 128512, 128591, 0, 0 };
  range["IsEmoticons"] = IsEmoticons;
  static const int IsEnclosedAlphanumericSupplement[] = { 127232, 127487, 0, 0 };
  range["IsEnclosedAlphanumericSupplement"] = IsEnclosedAlphanumericSupplement;
  static const int IsEnclosedAlphanumerics[] = { 9312, 9471, 0, 0 };
  range["IsEnclosedAlphanumerics"] = IsEnclosedAlphanumerics;
  static const int IsEnclosedCJKLettersandMonths[] = { 12800, 13055, 0, 0 };
  range["IsEnclosedCJKLettersandMonths"] = IsEnclosedCJKLettersandMonths;
  static const int IsEnclosedIdeographicSupplement[] = { 127488, 127743, 0, 0 };
  range["IsEnclosedIdeographicSupplement"] = IsEnclosedIdeographicSupplement;
  static const int IsEthiopic[] = { 4608, 4991, 0, 0 };
  range["IsEthiopic"] = IsEthiopic;
  static const int IsEthiopicExtended[] = { 11648, 11743, 0, 0 };
  range["IsEthiopicExtended"] = IsEthiopicExtended;
  static const int IsEthiopicExtended_A[] = { 43776, 43823, 0, 0 };
  range["IsEthiopicExtended-A"] = IsEthiopicExtended_A;
  static const int IsEthiopicExtended_B[] = { 124896, 124927, 0, 0 };
  range["IsEthiopicExtended-B"] = IsEthiopicExtended_B;
  static const int IsEthiopicSupplement[] = { 4992, 5023, 0, 0 };
  range["IsEthiopicSupplement"] = IsEthiopicSupplement;
  static const int IsGeneralPunctuation[] = { 8192, 8303, 0, 0 };
  range["IsGeneralPunctuation"] = IsGeneralPunctuation;
  static const int IsGeometricShapes[] = { 9632, 9727, 0, 0 };
  range["IsGeometricShapes"] = IsGeometricShapes;
  static const int IsGeometricShapesExtended[] = { 128896, 129023, 0, 0 };
  range["IsGeometricShapesExtended"] = IsGeometricShapesExtended;
  static const int IsGeorgian[] = { 4256, 4351, 0, 0 };
  range["IsGeorgian"] = IsGeorgian;
  static const int IsGeorgianExtended[] = { 7312, 7359, 0, 0 };
  range["IsGeorgianExtended"] = IsGeorgianExtended;
  static const int IsGeorgianSupplement[] = { 11520, 11567, 0, 0 };
  range["IsGeorgianSupplement"] = IsGeorgianSupplement;
  static const int IsGlagolitic[] = { 11264, 11359, 0, 0 };
  range["IsGlagolitic"] = IsGlagolitic;
  static const int IsGlagoliticSupplement[] = { 122880, 122927, 0, 0 };
  range["IsGlagoliticSupplement"] = IsGlagoliticSupplement;
  static const int IsGothic[] = { 66352, 66383, 0, 0 };
  range["IsGothic"] = IsGothic;
  static const int IsGrantha[] = { 70400, 70527, 0, 0 };
  range["IsGrantha"] = IsGrantha;
  static const int IsGreekExtended[] = { 7936, 8191, 0, 0 };
  range["IsGreekExtended"] = IsGreekExtended;
  static const int IsGreekandCoptic[] = { 880, 1023, 0, 0 };
  range["IsGreekandCoptic"] = IsGreekandCoptic;
  static const int IsGujarati[] = { 2688, 2815, 0, 0 };
  range["IsGujarati"] = IsGujarati;
  static const int IsGunjalaGondi[] = { 73056, 73135, 0, 0 };
  range["IsGunjalaGondi"] = IsGunjalaGondi;
  static const int IsGurmukhi[] = { 2560, 2687, 0, 0 };
  range["IsGurmukhi"] = IsGurmukhi;
  static const int IsHalfwidthandFullwidthForms[] = { 65280, 65519, 0, 0 };
  range["IsHalfwidthandFullwidthForms"] = IsHalfwidthandFullwidthForms;
  static const int IsHangulCompatibilityJamo[] = { 12592, 12687, 0, 0 };
  range["IsHangulCompatibilityJamo"] = IsHangulCompatibilityJamo;
  static const int IsHangulJamo[] = { 4352, 4607, 0, 0 };
  range["IsHangulJamo"] = IsHangulJamo;
  static const int IsHangulJamoExtended_A[] = { 43360, 43391, 0, 0 };
  range["IsHangulJamoExtended-A"] = IsHangulJamoExtended_A;
  static const int IsHangulJamoExtended_B[] = { 55216, 55295, 0, 0 };
  range["IsHangulJamoExtended-B"] = IsHangulJamoExtended_B;
  static const int IsHangulSyllables[] = { 44032, 55215, 0, 0 };
  range["IsHangulSyllables"] = IsHangulSyllables;
  static const int IsHanifiRohingya[] = { 68864, 68927, 0, 0 };
  range["IsHanifiRohingya"] = IsHanifiRohingya;
  static const int IsHanunoo[] = { 5920, 5951, 0, 0 };
  range["IsHanunoo"] = IsHanunoo;
  static const int IsHatran[] = { 67808, 67839, 0, 0 };
  range["IsHatran"] = IsHatran;
  static const int IsHebrew[] = { 1424, 1535, 0, 0 };
  range["IsHebrew"] = IsHebrew;
  static const int IsHighPrivateUseSurrogates[] = { 56192, 56319, 0, 0 };
  range["IsHighPrivateUseSurrogates"] = IsHighPrivateUseSurrogates;
  static const int IsHighSurrogates[] = { 55296, 56191, 0, 0 };
  range["IsHighSurrogates"] = IsHighSurrogates;
  static const int IsHiragana[] = { 12352, 12447, 0, 0 };
  range["IsHiragana"] = IsHiragana;
  static const int IsIPAExtensions[] = { 592, 687, 0, 0 };
  range["IsIPAExtensions"] = IsIPAExtensions;
  static const int IsIdeographicDescriptionCharacters[] = { 12272, 12287, 0, 0 };
  range["IsIdeographicDescriptionCharacters"] = IsIdeographicDescriptionCharacters;
  static const int IsIdeographicSymbolsandPunctuation[] = { 94176, 94207, 0, 0 };
  range["IsIdeographicSymbolsandPunctuation"] = IsIdeographicSymbolsandPunctuation;
  static const int IsImperialAramaic[] = { 67648, 67679, 0, 0 };
  range["IsImperialAramaic"] = IsImperialAramaic;
  static const int IsIndicSiyaqNumbers[] = { 126064, 126143, 0, 0 };
  range["IsIndicSiyaqNumbers"] = IsIndicSiyaqNumbers;
  static const int IsInscriptionalPahlavi[] = { 68448, 68479, 0, 0 };
  range["IsInscriptionalPahlavi"] = IsInscriptionalPahlavi;
  static const int IsInscriptionalParthian[] = { 68416, 68447, 0, 0 };
  range["IsInscriptionalParthian"] = IsInscriptionalParthian;
  static const int IsJavanese[] = { 43392, 43487, 0, 0 };
  range["IsJavanese"] = IsJavanese;
  static const int IsKaithi[] = { 69760, 69839, 0, 0 };
  range["IsKaithi"] = IsKaithi;
  static const int IsKaktovikNumerals[] = { 119488, 119519, 0, 0 };
  range["IsKaktovikNumerals"] = IsKaktovikNumerals;
  static const int IsKanaExtended_A[] = { 110848, 110895, 0, 0 };
  range["IsKanaExtended-A"] = IsKanaExtended_A;
  static const int IsKanaExtended_B[] = { 110576, 110591, 0, 0 };
  range["IsKanaExtended-B"] = IsKanaExtended_B;
  static const int IsKanaSupplement[] = { 110592, 110847, 0, 0 };
  range["IsKanaSupplement"] = IsKanaSupplement;
  static const int IsKanbun[] = { 12688, 12703, 0, 0 };
  range["IsKanbun"] = IsKanbun;
  static const int IsKangxiRadicals[] = { 12032, 12255, 0, 0 };
  range["IsKangxiRadicals"] = IsKangxiRadicals;
  static const int IsKannada[] = { 3200, 3327, 0, 0 };
  range["IsKannada"] = IsKannada;
  static const int IsKatakana[] = { 12448, 12543, 0, 0 };
  range["IsKatakana"] = IsKatakana;
  static const int IsKatakanaPhoneticExtensions[] = { 12784, 12799, 0, 0 };
  range["IsKatakanaPhoneticExtensions"] = IsKatakanaPhoneticExtensions;
  static const int IsKawi[] = { 73472, 73567, 0, 0 };
  range["IsKawi"] = IsKawi;
  static const int IsKayahLi[] = { 43264, 43311, 0, 0 };
  range["IsKayahLi"] = IsKayahLi;
  static const int IsKharoshthi[] = { 68096, 68191, 0, 0 };
  range["IsKharoshthi"] = IsKharoshthi;
  static const int IsKhitanSmallScript[] = { 101120, 101631, 0, 0 };
  range["IsKhitanSmallScript"] = IsKhitanSmallScript;
  static const int IsKhmer[] = { 6016, 6143, 0, 0 };
  range["IsKhmer"] = IsKhmer;
  static const int IsKhmerSymbols[] = { 6624, 6655, 0, 0 };
  range["IsKhmerSymbols"] = IsKhmerSymbols;
  static const int IsKhojki[] = { 70144, 70223, 0, 0 };
  range["IsKhojki"] = IsKhojki;
  static const int IsKhudawadi[] = { 70320, 70399, 0, 0 };
  range["IsKhudawadi"] = IsKhudawadi;
  static const int IsLao[] = { 3712, 3839, 0, 0 };
  range["IsLao"] = IsLao;
  static const int IsLatin_1Supplement[] = { 128, 255, 0, 0 };
  range["IsLatin-1Supplement"] = IsLatin_1Supplement;
  static const int IsLatinExtended_A[] = { 256, 383, 0, 0 };
  range["IsLatinExtended-A"] = IsLatinExtended_A;
  static const int IsLatinExtended_B[] = { 384, 591, 0, 0 };
  range["IsLatinExtended-B"] = IsLatinExtended_B;
  static const int IsLatinExtended_C[] = { 11360, 11391, 0, 0 };
  range["IsLatinExtended-C"] = IsLatinExtended_C;
  static const int IsLatinExtended_D[] = { 42784, 43007, 0, 0 };
  range["IsLatinExtended-D"] = IsLatinExtended_D;
  static const int IsLatinExtended_E[] = { 43824, 43887, 0, 0 };
  range["IsLatinExtended-E"] = IsLatinExtended_E;
  static const int IsLatinExtended_F[] = { 67456, 67519, 0, 0 };
  range["IsLatinExtended-F"] = IsLatinExtended_F;
  static const int IsLatinExtended_G[] = { 122624, 122879, 0, 0 };
  range["IsLatinExtended-G"] = IsLatinExtended_G;
  static const int IsLatinExtendedAdditional[] = { 7680, 7935, 0, 0 };
  range["IsLatinExtendedAdditional"] = IsLatinExtendedAdditional;
  static const int IsLepcha[] = { 7168, 7247, 0, 0 };
  range["IsLepcha"] = IsLepcha;
  static const int IsLetterlikeSymbols[] = { 8448, 8527, 0, 0 };
  range["IsLetterlikeSymbols"] = IsLetterlikeSymbols;
  static const int IsLimbu[] = { 6400, 6479, 0, 0 };
  range["IsLimbu"] = IsLimbu;
  static const int IsLinearA[] = { 67072, 67455, 0, 0 };
  range["IsLinearA"] = IsLinearA;
  static const int IsLinearBIdeograms[] = { 65664, 65791, 0, 0 };
  range["IsLinearBIdeograms"] = IsLinearBIdeograms;
  static const int IsLinearBSyllabary[] = { 65536, 65663, 0, 0 };
  range["IsLinearBSyllabary"] = IsLinearBSyllabary;
  static const int IsLisu[] = { 42192, 42239, 0, 0 };
  range["IsLisu"] = IsLisu;
  static const int IsLisuSupplement[] = { 73648, 73663, 0, 0 };
  range["IsLisuSupplement"] = IsLisuSupplement;
  static const int IsLowSurrogates[] = { 56320, 57343, 0, 0 };
  range["IsLowSurrogates"] = IsLowSurrogates;
  static const int IsLycian[] = { 66176, 66207, 0, 0 };
  range["IsLycian"] = IsLycian;
  static const int IsLydian[] = { 67872, 67903, 0, 0 };
  range["IsLydian"] = IsLydian;
  static const int IsMahajani[] = { 69968, 70015, 0, 0 };
  range["IsMahajani"] = IsMahajani;
  static const int IsMahjongTiles[] = { 126976, 127023, 0, 0 };
  range["IsMahjongTiles"] = IsMahjongTiles;
  static const int IsMakasar[] = { 73440, 73471, 0, 0 };
  range["IsMakasar"] = IsMakasar;
  static const int IsMalayalam[] = { 3328, 3455, 0, 0 };
  range["IsMalayalam"] = IsMalayalam;
  static const int IsMandaic[] = { 2112, 2143, 0, 0 };
  range["IsMandaic"] = IsMandaic;
  static const int IsManichaean[] = { 68288, 68351, 0, 0 };
  range["IsManichaean"] = IsManichaean;
  static const int IsMarchen[] = { 72816, 72895, 0, 0 };
  range["IsMarchen"] = IsMarchen;
  static const int IsMasaramGondi[] = { 72960, 73055, 0, 0 };
  range["IsMasaramGondi"] = IsMasaramGondi;
  static const int IsMathematicalAlphanumericSymbols[] = { 119808, 120831, 0, 0 };
  range["IsMathematicalAlphanumericSymbols"] = IsMathematicalAlphanumericSymbols;
  static const int IsMathematicalOperators[] = { 8704, 8959, 0, 0 };
  range["IsMathematicalOperators"] = IsMathematicalOperators;
  static const int IsMayanNumerals[] = { 119520, 119551, 0, 0 };
  range["IsMayanNumerals"] = IsMayanNumerals;
  static const int IsMedefaidrin[] = { 93760, 93855, 0, 0 };
  range["IsMedefaidrin"] = IsMedefaidrin;
  static const int IsMeeteiMayek[] = { 43968, 44031, 0, 0 };
  range["IsMeeteiMayek"] = IsMeeteiMayek;
  static const int IsMeeteiMayekExtensions[] = { 43744, 43775, 0, 0 };
  range["IsMeeteiMayekExtensions"] = IsMeeteiMayekExtensions;
  static const int IsMendeKikakui[] = { 124928, 125151, 0, 0 };
  range["IsMendeKikakui"] = IsMendeKikakui;
  static const int IsMeroiticCursive[] = { 68000, 68095, 0, 0 };
  range["IsMeroiticCursive"] = IsMeroiticCursive;
  static const int IsMeroiticHieroglyphs[] = { 67968, 67999, 0, 0 };
  range["IsMeroiticHieroglyphs"] = IsMeroiticHieroglyphs;
  static const int IsMiao[] = { 93952, 94111, 0, 0 };
  range["IsMiao"] = IsMiao;
  static const int IsMiscellaneousMathematicalSymbols_A[] = { 10176, 10223, 0, 0 };
  range["IsMiscellaneousMathematicalSymbols-A"] = IsMiscellaneousMathematicalSymbols_A;
  static const int IsMiscellaneousMathematicalSymbols_B[] = { 10624, 10751, 0, 0 };
  range["IsMiscellaneousMathematicalSymbols-B"] = IsMiscellaneousMathematicalSymbols_B;
  static const int IsMiscellaneousSymbols[] = { 9728, 9983, 0, 0 };
  range["IsMiscellaneousSymbols"] = IsMiscellaneousSymbols;
  static const int IsMiscellaneousSymbolsandArrows[] = { 11008, 11263, 0, 0 };
  range["IsMiscellaneousSymbolsandArrows"] = IsMiscellaneousSymbolsandArrows;
  static const int IsMiscellaneousSymbolsandPictographs[] = { 127744, 128511, 0, 0 };
  range["IsMiscellaneousSymbolsandPictographs"] = IsMiscellaneousSymbolsandPictographs;
  static const int IsMiscellaneousTechnical[] = { 8960, 9215, 0, 0 };
  range["IsMiscellaneousTechnical"] = IsMiscellaneousTechnical;
  static const int IsModi[] = { 71168, 71263, 0, 0 };
  range["IsModi"] = IsModi;
  static const int IsModifierToneLetters[] = { 42752, 42783, 0, 0 };
  range["IsModifierToneLetters"] = IsModifierToneLetters;
  static const int IsMongolian[] = { 6144, 6319, 0, 0 };
  range["IsMongolian"] = IsMongolian;
  static const int IsMongolianSupplement[] = { 71264, 71295, 0, 0 };
  range["IsMongolianSupplement"] = IsMongolianSupplement;
  static const int IsMro[] = { 92736, 92783, 0, 0 };
  range["IsMro"] = IsMro;
  static const int IsMultani[] = { 70272, 70319, 0, 0 };
  range["IsMultani"] = IsMultani;
  static const int IsMusicalSymbols[] = { 119040, 119295, 0, 0 };
  range["IsMusicalSymbols"] = IsMusicalSymbols;
  static const int IsMyanmar[] = { 4096, 4255, 0, 0 };
  range["IsMyanmar"] = IsMyanmar;
  static const int IsMyanmarExtended_A[] = { 43616, 43647, 0, 0 };
  range["IsMyanmarExtended-A"] = IsMyanmarExtended_A;
  static const int IsMyanmarExtended_B[] = { 43488, 43519, 0, 0 };
  range["IsMyanmarExtended-B"] = IsMyanmarExtended_B;
  static const int IsNKo[] = { 1984, 2047, 0, 0 };
  range["IsNKo"] = IsNKo;
  static const int IsNabataean[] = { 67712, 67759, 0, 0 };
  range["IsNabataean"] = IsNabataean;
  static const int IsNagMundari[] = { 124112, 124159, 0, 0 };
  range["IsNagMundari"] = IsNagMundari;
  static const int IsNandinagari[] = { 72096, 72191, 0, 0 };
  range["IsNandinagari"] = IsNandinagari;
  static const int IsNewTaiLue[] = { 6528, 6623, 0, 0 };
  range["IsNewTaiLue"] = IsNewTaiLue;
  static const int IsNewa[] = { 70656, 70783, 0, 0 };
  range["IsNewa"] = IsNewa;
  static const int IsNumberForms[] = { 8528, 8591, 0, 0 };
  range["IsNumberForms"] = IsNumberForms;
  static const int IsNushu[] = { 110960, 111359, 0, 0 };
  range["IsNushu"] = IsNushu;
  static const int IsNyiakengPuachueHmong[] = { 123136, 123215, 0, 0 };
  range["IsNyiakengPuachueHmong"] = IsNyiakengPuachueHmong;
  static const int IsOgham[] = { 5760, 5791, 0, 0 };
  range["IsOgham"] = IsOgham;
  static const int IsOlChiki[] = { 7248, 7295, 0, 0 };
  range["IsOlChiki"] = IsOlChiki;
  static const int IsOldHungarian[] = { 68736, 68863, 0, 0 };
  range["IsOldHungarian"] = IsOldHungarian;
  static const int IsOldItalic[] = { 66304, 66351, 0, 0 };
  range["IsOldItalic"] = IsOldItalic;
  static const int IsOldNorthArabian[] = { 68224, 68255, 0, 0 };
  range["IsOldNorthArabian"] = IsOldNorthArabian;
  static const int IsOldPermic[] = { 66384, 66431, 0, 0 };
  range["IsOldPermic"] = IsOldPermic;
  static const int IsOldPersian[] = { 66464, 66527, 0, 0 };
  range["IsOldPersian"] = IsOldPersian;
  static const int IsOldSogdian[] = { 69376, 69423, 0, 0 };
  range["IsOldSogdian"] = IsOldSogdian;
  static const int IsOldSouthArabian[] = { 68192, 68223, 0, 0 };
  range["IsOldSouthArabian"] = IsOldSouthArabian;
  static const int IsOldTurkic[] = { 68608, 68687, 0, 0 };
  range["IsOldTurkic"] = IsOldTurkic;
  static const int IsOldUyghur[] = { 69488, 69551, 0, 0 };
  range["IsOldUyghur"] = IsOldUyghur;
  static const int IsOpticalCharacterRecognition[] = { 9280, 9311, 0, 0 };
  range["IsOpticalCharacterRecognition"] = IsOpticalCharacterRecognition;
  static const int IsOriya[] = { 2816, 2943, 0, 0 };
  range["IsOriya"] = IsOriya;
  static const int IsOrnamentalDingbats[] = { 128592, 128639, 0, 0 };
  range["IsOrnamentalDingbats"] = IsOrnamentalDingbats;
  static const int IsOsage[] = { 66736, 66815, 0, 0 };
  range["IsOsage"] = IsOsage;
  static const int IsOsmanya[] = { 66688, 66735, 0, 0 };
  range["IsOsmanya"] = IsOsmanya;
  static const int IsOttomanSiyaqNumbers[] = { 126208, 126287, 0, 0 };
  range["IsOttomanSiyaqNumbers"] = IsOttomanSiyaqNumbers;
  static const int IsPahawhHmong[] = { 92928, 93071, 0, 0 };
  range["IsPahawhHmong"] = IsPahawhHmong;
  static const int IsPalmyrene[] = { 67680, 67711, 0, 0 };
  range["IsPalmyrene"] = IsPalmyrene;
  static const int IsPauCinHau[] = { 72384, 72447, 0, 0 };
  range["IsPauCinHau"] = IsPauCinHau;
  static const int IsPhags_pa[] = { 43072, 43135, 0, 0 };
  range["IsPhags-pa"] = IsPhags_pa;
  static const int IsPhaistosDisc[] = { 66000, 66047, 0, 0 };
  range["IsPhaistosDisc"] = IsPhaistosDisc;
  static const int IsPhoenician[] = { 67840, 67871, 0, 0 };
  range["IsPhoenician"] = IsPhoenician;
  static const int IsPhoneticExtensions[] = { 7424, 7551, 0, 0 };
  range["IsPhoneticExtensions"] = IsPhoneticExtensions;
  static const int IsPhoneticExtensionsSupplement[] = { 7552, 7615, 0, 0 };
  range["IsPhoneticExtensionsSupplement"] = IsPhoneticExtensionsSupplement;
  static const int IsPlayingCards[] = { 127136, 127231, 0, 0 };
  range["IsPlayingCards"] = IsPlayingCards;
  static const int IsPrivateUseArea[] = { 57344, 63743, 0, 0 };
  range["IsPrivateUseArea"] = IsPrivateUseArea;
  static const int IsPsalterPahlavi[] = { 68480, 68527, 0, 0 };
  range["IsPsalterPahlavi"] = IsPsalterPahlavi;
  static const int IsRejang[] = { 43312, 43359, 0, 0 };
  range["IsRejang"] = IsRejang;
  static const int IsRumiNumeralSymbols[] = { 69216, 69247, 0, 0 };
  range["IsRumiNumeralSymbols"] = IsRumiNumeralSymbols;
  static const int IsRunic[] = { 5792, 5887, 0, 0 };
  range["IsRunic"] = IsRunic;
  static const int IsSamaritan[] = { 2048, 2111, 0, 0 };
  range["IsSamaritan"] = IsSamaritan;
  static const int IsSaurashtra[] = { 43136, 43231, 0, 0 };
  range["IsSaurashtra"] = IsSaurashtra;
  static const int IsSharada[] = { 70016, 70111, 0, 0 };
  range["IsSharada"] = IsSharada;
  static const int IsShavian[] = { 66640, 66687, 0, 0 };
  range["IsShavian"] = IsShavian;
  static const int IsShorthandFormatControls[] = { 113824, 113839, 0, 0 };
  range["IsShorthandFormatControls"] = IsShorthandFormatControls;
  static const int IsSiddham[] = { 71040, 71167, 0, 0 };
  range["IsSiddham"] = IsSiddham;
  static const int IsSinhala[] = { 3456, 3583, 0, 0 };
  range["IsSinhala"] = IsSinhala;
  static const int IsSinhalaArchaicNumbers[] = { 70112, 70143, 0, 0 };
  range["IsSinhalaArchaicNumbers"] = IsSinhalaArchaicNumbers;
  static const int IsSmallFormVariants[] = { 65104, 65135, 0, 0 };
  range["IsSmallFormVariants"] = IsSmallFormVariants;
  static const int IsSmallKanaExtension[] = { 110896, 110959, 0, 0 };
  range["IsSmallKanaExtension"] = IsSmallKanaExtension;
  static const int IsSogdian[] = { 69424, 69487, 0, 0 };
  range["IsSogdian"] = IsSogdian;
  static const int IsSoraSompeng[] = { 69840, 69887, 0, 0 };
  range["IsSoraSompeng"] = IsSoraSompeng;
  static const int IsSoyombo[] = { 72272, 72367, 0, 0 };
  range["IsSoyombo"] = IsSoyombo;
  static const int IsSpacingModifierLetters[] = { 688, 767, 0, 0 };
  range["IsSpacingModifierLetters"] = IsSpacingModifierLetters;
  static const int IsSpecials[] = { 65520, 65535, 0, 0 };
  range["IsSpecials"] = IsSpecials;
  static const int IsSundanese[] = { 7040, 7103, 0, 0 };
  range["IsSundanese"] = IsSundanese;
  static const int IsSundaneseSupplement[] = { 7360, 7375, 0, 0 };
  range["IsSundaneseSupplement"] = IsSundaneseSupplement;
  static const int IsSuperscriptsandSubscripts[] = { 8304, 8351, 0, 0 };
  range["IsSuperscriptsandSubscripts"] = IsSuperscriptsandSubscripts;
  static const int IsSupplementalArrows_A[] = { 10224, 10239, 0, 0 };
  range["IsSupplementalArrows-A"] = IsSupplementalArrows_A;
  static const int IsSupplementalArrows_B[] = { 10496, 10623, 0, 0 };
  range["IsSupplementalArrows-B"] = IsSupplementalArrows_B;
  static const int IsSupplementalArrows_C[] = { 129024, 129279, 0, 0 };
  range["IsSupplementalArrows-C"] = IsSupplementalArrows_C;
  static const int IsSupplementalMathematicalOperators[] = { 10752, 11007, 0, 0 };
  range["IsSupplementalMathematicalOperators"] = IsSupplementalMathematicalOperators;
  static const int IsSupplementalPunctuation[] = { 11776, 11903, 0, 0 };
  range["IsSupplementalPunctuation"] = IsSupplementalPunctuation;
  static const int IsSupplementalSymbolsandPictographs[] = { 129280, 129535, 0, 0 };
  range["IsSupplementalSymbolsandPictographs"] = IsSupplementalSymbolsandPictographs;
  static const int IsSupplementaryPrivateUseArea_A[] = { 983040, 1048575, 0, 0 };
  range["IsSupplementaryPrivateUseArea-A"] = IsSupplementaryPrivateUseArea_A;
  static const int IsSupplementaryPrivateUseArea_B[] = { 1048576, 1114111, 0, 0 };
  range["IsSupplementaryPrivateUseArea-B"] = IsSupplementaryPrivateUseArea_B;
  static const int IsSuttonSignWriting[] = { 120832, 121519, 0, 0 };
  range["IsSuttonSignWriting"] = IsSuttonSignWriting;
  static const int IsSylotiNagri[] = { 43008, 43055, 0, 0 };
  range["IsSylotiNagri"] = IsSylotiNagri;
  static const int IsSymbolsandPictographsExtended_A[] = { 129648, 129791, 0, 0 };
  range["IsSymbolsandPictographsExtended-A"] = IsSymbolsandPictographsExtended_A;
  static const int IsSymbolsforLegacyComputing[] = { 129792, 130047, 0, 0 };
  range["IsSymbolsforLegacyComputing"] = IsSymbolsforLegacyComputing;
  static const int IsSyriac[] = { 1792, 1871, 0, 0 };
  range["IsSyriac"] = IsSyriac;
  static const int IsSyriacSupplement[] = { 2144, 2159, 0, 0 };
  range["IsSyriacSupplement"] = IsSyriacSupplement;
  static const int IsTagalog[] = { 5888, 5919, 0, 0 };
  range["IsTagalog"] = IsTagalog;
  static const int IsTagbanwa[] = { 5984, 6015, 0, 0 };
  range["IsTagbanwa"] = IsTagbanwa;
  static const int IsTags[] = { 917504, 917631, 0, 0 };
  range["IsTags"] = IsTags;
  static const int IsTaiLe[] = { 6480, 6527, 0, 0 };
  range["IsTaiLe"] = IsTaiLe;
  static const int IsTaiTham[] = { 6688, 6831, 0, 0 };
  range["IsTaiTham"] = IsTaiTham;
  static const int IsTaiViet[] = { 43648, 43743, 0, 0 };
  range["IsTaiViet"] = IsTaiViet;
  static const int IsTaiXuanJingSymbols[] = { 119552, 119647, 0, 0 };
  range["IsTaiXuanJingSymbols"] = IsTaiXuanJingSymbols;
  static const int IsTakri[] = { 71296, 71375, 0, 0 };
  range["IsTakri"] = IsTakri;
  static const int IsTamil[] = { 2944, 3071, 0, 0 };
  range["IsTamil"] = IsTamil;
  static const int IsTamilSupplement[] = { 73664, 73727, 0, 0 };
  range["IsTamilSupplement"] = IsTamilSupplement;
  static const int IsTangsa[] = { 92784, 92879, 0, 0 };
  range["IsTangsa"] = IsTangsa;
  static const int IsTangut[] = { 94208, 100351, 0, 0 };
  range["IsTangut"] = IsTangut;
  static const int IsTangutComponents[] = { 100352, 101119, 0, 0 };
  range["IsTangutComponents"] = IsTangutComponents;
  static const int IsTangutSupplement[] = { 101632, 101759, 0, 0 };
  range["IsTangutSupplement"] = IsTangutSupplement;
  static const int IsTelugu[] = { 3072, 3199, 0, 0 };
  range["IsTelugu"] = IsTelugu;
  static const int IsThaana[] = { 1920, 1983, 0, 0 };
  range["IsThaana"] = IsThaana;
  static const int IsThai[] = { 3584, 3711, 0, 0 };
  range["IsThai"] = IsThai;
  static const int IsTibetan[] = { 3840, 4095, 0, 0 };
  range["IsTibetan"] = IsTibetan;
  static const int IsTifinagh[] = { 11568, 11647, 0, 0 };
  range["IsTifinagh"] = IsTifinagh;
  static const int IsTirhuta[] = { 70784, 70879, 0, 0 };
  range["IsTirhuta"] = IsTirhuta;
  static const int IsToto[] = { 123536, 123583, 0, 0 };
  range["IsToto"] = IsToto;
  static const int IsTransportandMapSymbols[] = { 128640, 128767, 0, 0 };
  range["IsTransportandMapSymbols"] = IsTransportandMapSymbols;
  static const int IsUgaritic[] = { 66432, 66463, 0, 0 };
  range["IsUgaritic"] = IsUgaritic;
  static const int IsUnifiedCanadianAboriginalSyllabics[] = { 5120, 5759, 0, 0 };
  range["IsUnifiedCanadianAboriginalSyllabics"] = IsUnifiedCanadianAboriginalSyllabics;
  static const int IsUnifiedCanadianAboriginalSyllabicsExtended[] = { 6320, 6399, 0, 0 };
  range["IsUnifiedCanadianAboriginalSyllabicsExtended"] = IsUnifiedCanadianAboriginalSyllabicsExtended;
  static const int IsUnifiedCanadianAboriginalSyllabicsExtended_A[] = { 72368, 72383, 0, 0 };
  range["IsUnifiedCanadianAboriginalSyllabicsExtended-A"] = IsUnifiedCanadianAboriginalSyllabicsExtended_A;
  static const int IsVai[] = { 42240, 42559, 0, 0 };
  range["IsVai"] = IsVai;
  static const int IsVariationSelectors[] = { 65024, 65039, 0, 0 };
  range["IsVariationSelectors"] = IsVariationSelectors;
  static const int IsVariationSelectorsSupplement[] = { 917760, 917999, 0, 0 };
  range["IsVariationSelectorsSupplement"] = IsVariationSelectorsSupplement;
  static const int IsVedicExtensions[] = { 7376, 7423, 0, 0 };
  range["IsVedicExtensions"] = IsVedicExtensions;
  static const int IsVerticalForms[] = { 65040, 65055, 0, 0 };
  range["IsVerticalForms"] = IsVerticalForms;
  static const int IsVithkuqi[] = { 66928, 67007, 0, 0 };
  range["IsVithkuqi"] = IsVithkuqi;
  static const int IsWancho[] = { 123584, 123647, 0, 0 };
  range["IsWancho"] = IsWancho;
  static const int IsWarangCiti[] = { 71840, 71935, 0, 0 };
  range["IsWarangCiti"] = IsWarangCiti;
  static const int IsYezidi[] = { 69248, 69311, 0, 0 };
  range["IsYezidi"] = IsYezidi;
  static const int IsYiRadicals[] = { 42128, 42191, 0, 0 };
  range["IsYiRadicals"] = IsYiRadicals;
  static const int IsYiSyllables[] = { 40960, 42127, 0, 0 };
  range["IsYiSyllables"] = IsYiSyllables;
  static const int IsYijingHexagramSymbols[] = { 19904, 19967, 0, 0 };
  range["IsYijingHexagramSymbols"] = IsYijingHexagramSymbols;
  static const int IsZanabazarSquare[] = { 72192, 72271, 0, 0 };
  range["IsZanabazarSquare"] = IsZanabazarSquare;
  static const int IsZnamennyMusicalNotation[] = { 118528, 118735, 0, 0 };
  range["IsZnamennyMusicalNotation"] = IsZnamennyMusicalNotation;
}
